Require Import ssreflect ssrbool ssrfun eqtype ssrnat seq choice fintype.
Require Import div finfun bigop prime binomial ssralg finset fingroup finalg.
Require Import perm zmodp matrix.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Local Open Scope ring_scope. 

Section permseq_defs.

Variable n : nat.

Definition seqnat := seq nat.

(* Transformation from a permutation to a sequence *)

Definition seq_of_perm (s:'S_n) : seqnat :=
  map (fun i => val(s i)) (enum 'I_n).

Definition fun_of_permseq (s : seqnat) := fun i => (nth 0%N s i).

Coercion fun_of_permseq : seqnat >-> Funclass.

Lemma fun_of_permseqE (s1 : seqnat) (i:'I_n) : s1 i = (nth 0%N s1 (val i)).
Proof. by []. Qed.

Lemma seq_of_permE : forall (s:'S_n) (i: 'I_n), (seq_of_perm s) i = s i.
Proof.
move => s0 i. rewrite /fun_of_permseq.
rewrite (nth_map i); last by rewrite size_enum_ord ltn_ord.
have -> : (nth i (enum 'I_n) i) = i; last by done.
by apply : (ord_inj (nth_enum_ord i (ltn_ord i))).
Qed.

(* Transformation from sequence to a permutation *)

Variable s : seqnat.

Hypothesis seqperm : perm_eq s (iota 0 n).


Lemma seqperm_dim x : x < n -> x\in s.
Proof.
move : (perm_eq_mem seqperm) => H1 H2.
by rewrite (H1 x) mem_iota add0n.
Qed.

Lemma seqperm_dim2 x : x\in s ->  x < n.
Proof.
move : (perm_eq_mem seqperm) => H1 H2.
by rewrite (H1 x) mem_iota add0n in H2. 
Qed.


Lemma sperm_ord_proof (i:'I_n) : s i < n.
Proof.
have H : n = size s; first by rewrite (perm_eq_size seqperm) size_iota.
have H1 : i < size s; first by rewrite -H (ltn_ord i).
have H2 : s i \in s; first by rewrite (mem_nth 0%N H1).
by rewrite (seqperm_dim2 H2).
Qed.


Definition f_of_s (i:'I_n) := Ordinal (sperm_ord_proof i).

Lemma injective_f_of_s : injective f_of_s.
Proof.
move => x1 x2 H.
have H0 : n = size s; first by rewrite (perm_eq_size seqperm) size_iota.
have H1 : val (f_of_s x1) =  (nth 0%N s (val x1)); first by done.
have H2 : val (f_of_s x2) =  (nth 0%N s (val x2)); first by done.
have HH : (nth 0%N s (val x1)) = (nth 0%N s (val x2)); first by rewrite -H1 -H2 H.
have H3 : (val x1) < (size s); first by rewrite -H0 /= (ltn_ord x1).
have H4 : (val x2) < (size s); first by rewrite -H0 /= (ltn_ord x2).
have H5 : (uniq s); first by rewrite (perm_eq_uniq seqperm) iota_uniq.
move : HH. move/eqP.
rewrite (nth_uniq 0%N H3 H4 H5). move/eqP.
by apply : val_inj.
Qed.

Definition perm_of_seq (s : seqnat) : 'S_n := (perm injective_f_of_s).

Lemma perm_of_seqE (i:'I_n) : s i = (perm_of_seq s) i.
Proof.
by rewrite permE.
Qed.

End permseq_defs.


Section permseq_lemmas.

Variable n:nat.


Lemma eq_nat_of_ord_eq_ord2  (i j:'I_n) : (nat_of_ord i)=(nat_of_ord j) -> i=j.
Proof.
by move/eqP; move=>H; apply : val_inj; move : (eqP H) => H1.
Qed.

Lemma seq_of_perm_mem (s:'S_n) : forall x, x \in seq_of_perm s -> x < n.
Proof.
move=> x.
move/mapP => [y H]; rewrite /= => ->; apply : (ltn_ord (s y)).
Qed.

Lemma seq_of_perm_mem_inv (s:'S_n) : forall x, (x < n) = false -> (x \in seq_of_perm s) = false.
Proof.
move=> x H.
by apply/negP => H1; move : (seq_of_perm_mem H1); rewrite H.
Qed.

Lemma index_mem_eq_size (T:eqType) (i:T) (s : seq T) : index i s = size s -> i \notin s. 
Proof.
move => H.
by apply/negP; rewrite -index_mem H ltnn.
Qed.

Lemma seq_of_perm_uniq (s:'S_n) : uniq (seq_of_perm s).
Proof.
rewrite map_inj_in_uniq; first by rewrite enum_uniq.
move => x y Hx Hy /= => H.
apply : (@perm_inj (ordinal_finType n) s x y (eq_nat_of_ord_eq_ord2 H)).
Qed.

Lemma size_seq_of_perm (s:'S_n) : n = size  (seq_of_perm s). 
Proof.
by rewrite size_map size_enum_ord.
Qed.

Lemma perm_eq_seq_of_perm_iota (s:'S_n) : perm_eq (seq_of_perm s) (iota 0 n).
Proof.
apply : uniq_perm_eq.
  - by apply : seq_of_perm_uniq.
  - by apply : iota_uniq.
  - move => x. rewrite mem_iota /= add0n.
    case H : (x < n); last by apply : (seq_of_perm_mem_inv s H).
    apply/mapP.
    exists (s^-1 (Ordinal H))%g.
    by rewrite mem_enum inE.
    by rewrite permKV.
Qed.    

Lemma seq_of_permP (s : seqnat) (s1 : 'S_n) :
  [/\ n = size s, perm_eq s (iota 0 n) &
  forall (i:'I_n), s i = s1 i] <-> s = seq_of_perm s1.
Proof.
split=> [[Hm Hn HMN]|->].
  have H1 : size s = size (seq_of_perm s1); first by rewrite size_map size_enum_ord Hm.
  have H2 : (forall i : nat, i < size s -> nth 0%N s i = nth 0%N (seq_of_perm s1) i).
   move => i H. rewrite -Hm in H. move : (HMN (Ordinal H)).
   rewrite /= /fun_of_permseq  => ->.
   have -> : nth 0%N (seq_of_perm s1) i = (seq_of_perm s1) i; first by rewrite /fun_of_permseq.
   by rewrite -seq_of_permE.
  by apply : (eq_from_nth H1 H2).
split; first by rewrite size_map size_enum_ord.
  by apply : (perm_eq_seq_of_perm_iota s1).
by move => i; rewrite seq_of_permE.
Qed.

Section perm_of_seqP. 

Variable s : seqnat.
Variable s1 : 'S_n.
Hypothesis H1 : n = size s.
Hypothesis H2 : perm_eq s (iota 0 n).


Lemma perm_of_seqP : (perm_of_seq H2 s) = s1 <-> s = seq_of_perm s1.
Proof.
split.
  rewrite -seq_of_permP.
  split => //=.
  by rewrite -H => i; rewrite perm_of_seqE.
rewrite -seq_of_permP.
case => Ha Hb Hc.
apply/permP => i.
have H : (nat_of_ord ((perm_of_seq H2 s) i)) = (nat_of_ord (s1 i)).
  by rewrite -Hc (perm_of_seqE H2 i).
by apply : (eq_nat_of_ord_eq_ord2 H).
Qed.

End perm_of_seqP. 


Lemma perm_of_seq_seq_of_perm (s1:'S_n) : (perm_of_seq (perm_eq_seq_of_perm_iota s1) (seq_of_perm s1)) = s1.
Proof.
by rewrite (@perm_of_seqP (seq_of_perm s1) s1 (size_seq_of_perm s1) (perm_eq_seq_of_perm_iota s1)).
Qed.

Variable s : seqnat.
Hypothesis H1 : n = size s.
Hypothesis H2 : perm_eq s (iota 0 n).

Lemma seq_of_perm_perm_of_seq : s = (seq_of_perm (perm_of_seq H2 s)).
Proof. 
by rewrite -(@perm_of_seqP s (perm_of_seq H2 s) H1 H2).
Qed.

End permseq_lemmas.



