Add LoadPath "CoqEAL".

Require Import ssreflect ssrbool ssrfun eqtype ssrnat seq div choice fintype.
Require Import finfun finset ssralg perm zmodp seqmatrix1 seqmatrix_adds advf_type.
Require Import Recdef.
Require Import invmx.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.


(* Main functions:

genDvf M: compute an admissible discrete vector field from M (no ordered)
genOrders M: returns the relations created when a vector field is built
dvford M: compute an admissible ordered discrete vector field from M
reorderM M dvf: the matrix M is reordered taking into account dvf
inverse M: given a triangular inferior matrix, the inverse is computed
matrixReduced M: given a matrix compute a reduced matrix*)



Section definitions.

Local Open Scope ring_scope.


(*check that every vector (c,d) of vf satisfies that the entry M(c,d) is 1*)
Definition compijCvd (vf:vectorfield)(M: matZ2) := 
  all [pred p| compij p.1 p.2 M == 1] vf.



Section matrices.

(*-- get the column c of the matrix given*)
Fixpoint getcol (c:nat)(s: matZ2) :=
match s with
|nil => nil
|a::b =>  ((nth 0 a c)::nil) ++ (getcol c b)
end.


Definition vectorZeros n := nseq n (0:Z2).

Definition matrizZeros m n := nseq m (vectorZeros n).




(*Operations with matrices*)

(*Working in Z2 the addition gives the same result than the substraction*)

(*mulseqmx s1 s2 :  Product of two matrices (sequences) is: mulseqmx s1 s2*)
(*addseqmx s1 s2 :  Addition of two matrices (sequences)*)
(* subseqmx s1 s2 :  Substraction of two matrices (sequences)*)
(*invmx_seqmx n s : the inverse of a matrix is computed with the function invmx_seqmx. It receives two parameters: the size of the square matrix and the matrix (sequences)*)

End matrices.



Section vectors.

Section seqcvd.



(* return the sequences which finish with k *)
Definition getlastE (k:nat)(ord: orders):= filter [pred s | last 0%N s == k] ord.


Fixpoint firstElem1 k (s:seqZ2):nat:=
match s with
|nil => 0%N  
|a::b => match k with
          |0 => if (a!=0) then 0%N else (1%N + (firstElem1 0 b))%N
          |S p => (1%N + (firstElem1 (k.-1) b))%N
          end
end.


Functional Scheme firstElem1_ind := Induction for firstElem1 Sort Prop.


End seqcvd.

Local Close Scope ring_scope.

(*DVF*)

(*-- The vector that we have to add, has to very that the first component is less or igual than 
 -- the number of rows and the second one than the number of columns*)
Definition verifyLength (m n a b: nat):= ((0<=a) && (a<m) && (0<=b) && (b<n)).


(*-- Check if the vector (a,b) can be added, if it hasn't any vectors that starts with a or 
  -- finishes with b return true in other case false *)
(*notin*)
Definition canAddCvd (m n a b: nat)(vf: vectorfield):=
 (((a \in (getfirstElemseq vf))== false) && ((b \in (getsndElemseq vf))== false)
     && (verifyLength m n a b)).


(*-- return a list with pairs with the given list and the new vector (a,b) *)
Definition addcvd (a b: nat)(vf:vectorfield):= ((a, b)::nil) ++ vf.



(*check that the order i > j can be added and check that will not generate cycle with a sequence rest which is a partial order*)
(*if i or j are not in rest, there will not be problem; but i and j are in rest and the position i is higher than the position j will be generated a cycle*)
(*to check canAddOrder*)


Definition nocycles_order  (i j:nat)(ord: order):=
match ord with
|nil => true
|a::b => (((i \in (a::b))==false)  || ((j \in (a::b))==false) ||
           ((index i (a::b)) < (index j (a::b))))
end.


(*check any sequence of rest will generate cycles with the partial order i > j*)
Fixpoint canAddOrder (i j:nat)(ords:orders):=
match ords with
|nil => (i!=j)
|a::b => if (i!=j)
            then if ((nocycles_order i j a) == true)
		then (canAddOrder i j b)
		else false
            else false
end.




(*to check canAOrder, check if cycles will be generated when this new order can join several orders *)
(*l1 are the sequences which finish with j, l2 are the sequences which start with i*)
(*check that in the concatenation of every sequence of l1 with every sequence of l2 
does not generate cycles or there will no be repeated in the new sequence*)
Fixpoint nocycles_concat_order (l1:order) (l2:orders) :=
match l1, l2 with
|_, nil => true
|nil, _ => true
|a::b, c::d => if uniq (c ++ (a::b))
                    then (nocycles_concat_order l1 d)
                    else false
end.


(* return True if there are not cycles*)
Fixpoint nocycles_concat_orders (l1 l2:orders) := 
match l1, l2 with
|nil, _ => true
|a::b ,nil => true
|a::b, c::d => if (nocycles_concat_order a (c::d))
                  then (nocycles_concat_orders b (c::d))
                  else false
end.



(*--check if the restriction i < j will generate cycles joining several previous restrictions.
-- we us cycles2 with the orders start with i and the orders finish with j*)
Definition canAOrder (i j:nat) (s: orders) :=
match s with
|nil => true
|a::b => let getl := (getlastE i (a::b)) in
          let getf := (getfirstE j (a::b)) in
          if ((getf == nil) || (getl == nil))
          then true
          else (nocycles_concat_orders (getf)(getl))
end. 


(*ADD ORDERS*)
(*-- It's necessary to prove that (canAddOrder, canAOrder = True)
-- addOrder: add the order [i,k] in the matrix of orders
-- It's necessary to add other orders. 
-- if some of them is of the way [h,i] we need add the order [i,k] and [h,i,k] or
-- if some of them is of the way [k,j] we need add the order [i,k] and [i,k,j]*)

(*this function creates if it is possible a new order by the beginning or the end*)
Definition addOrder_order (i j:nat)(r:order):=
match r with
|nil => nil 
|a::b => if (i == (last 0%N (a::b)))
           then if (j == (head 0%N  (a::b)))
                  then nil
                  else (((a::b) ++ (j::nil))::nil)
           else if (j == (head 0%N  (a::b)))
                  then (((i::nil)++(a::b))::nil)
                  else nil 
end.

(*add new orders which will be generated when you add i < j
These orders will be concatenation of a sequence of rest with the new order i < j*)
Fixpoint addOrder_orders (i j : nat)(ords: orders):=
match ords with
|nil =>  nil
|a::b => (a::nil)++ if (i!=j)
                      then (addOrder_order i j a) ++ (addOrder_orders i j b)
                      else nil
end.



(*We add an iterator so that the function ends in some momenta*)
(*-- add more restrictions, for instance: [l,i],[k,o] -> [l,i],[k,o],[l,i,k,o]*)

(*add new orders taking a sequence of r1 with other of r2*)

Function addOrder_concat (ords: orders * orders){measure (fun ords => ((size (fst ords)) + (size (snd ords)))%N)}:orders:=
match (fst ords), (snd ords) with
| nil , _ => nil
| _ , nil => nil
|a::nil, c::d => ((c++a)::nil) ++ (addOrder_concat ((a::nil), d))
|a::b, c::nil =>  ((c++a)::nil) ++ (addOrder_concat (b,(c::nil)))
|a::b , c::d => ((c++a)::nil) ++ (addOrder_concat ((a::nil), d)) ++ (addOrder_concat (b, (c::d)))
end.
Proof.
+ rewrite /=; move=> ords a b.
  move=> _ -> c d -> //=.
+ rewrite /=; move=> ords a b s s0. 
  move=> _ -> c d _ -> //=.
+ rewrite /=; move=> ords a b s s0. 
  move=> _ -> c d s1 s2 _ -> //=.
+ rewrite /=; move=> ords a b s s0. 
  move=> _ -> c d s1 s2 _ -> //=.
  rewrite add1n -{1}(add0n ((size s2).+2)).
  apply/ltP.
  by rewrite ltn_add2r.
Defined.



(*-- It's necessary to prove that (canAddOrder canAOrder = True)
-- i means the row (or the first pair) of the vector to add
-- the second argument is always 1 (iterator)
-- the third argument is the column j-esima since we need add the vector [(i,j)]
-- and the correspondent orders. These orders would be some of these [(i,1),..,(i,i-1),(i,i+1),..,(i,numcolumnas)] 
*)

Local Open Scope ring_scope.

(*add every orders which are generated when you select a new vector (a,b) to add it to the vector field,
 it is necessary to look in the elements of the column b*)
Fixpoint addOrders (i k:nat)(s:seqZ2)(ords: orders):=
match s, ords with
|nil, _ => ords
|a::nil, nil => if ((a!= 0) && (i!=k)) 
                   then (i::k::nil)::nil 
                   else nil
|a::nil, _ => let getf := (getfirstE k ords) in
                 let getl := (getlastE i ords) in
                 if ((a!=0) && (i!=k)) 
                      then if ((getf == nil) || (getl == nil))
                             then ((i::k::nil)::nil ) ++ (addOrder_orders i k ords)
                             else ((i::k::nil)::nil ) ++(addOrder_orders i k ords)++(addOrder_concat (getf, getl))  
                   else ords
|a::b, nil => if ((a != 0) && (i != k))
		 then addOrders i (k+1) b ((i::k::nil)::nil)
		else addOrders i (k+1) b nil
|a::b, _ => let getf := (getfirstE k ords) in
                 let getl := (getlastE i ords) in
                 if ((a != 0) && (i != k))
                      then (addOrders i (k+1) b (((i::k::nil)::nil) ++ (addOrder_orders i k ords) ++ (addOrder_concat (getf, getl))))
		  else addOrders i (k+1) b ords
end.

Functional Scheme addOrders_ind := Induction for addOrders Sort Prop. 


(*Check if it is possible adding the new order i > k taking into account the column and the others partial orders*)
Fixpoint canAddOrders (i k:nat)(s:seqZ2)(ord: orders) :=
match s, ord with
|nil, _ => true
|a::b, nil => if ((i!=k) && (a != 0))
               then (canAddOrders i (k+1) b ((i::k::nil)::nil)) 
               else (canAddOrders i (k+1) b nil)
|a::b, c::d => if ((i!=k) &&(a!= 0))
                   then if ((canAddOrder i k (c::d)) && (canAOrder i k (c::d)))
			then (canAddOrders i (k+1) b (((i::k::nil)::nil) ++ (addOrder_orders i k (c::d)) 
                                 ++  (addOrder_concat ((getfirstE k ord), (getlastE i ord)))))   
		        else false
		 else (canAddOrders i (k+1) b (c::d))
end.
  

Functional Scheme canAddOrders_ind := Induction for canAddOrders Sort Prop.
       


(*it = m*n+1*)
(*returns a pair where the first component is the dvf of mm and the second one is the partial orders*)
Fixpoint genDvfOrders (it:nat)(i j:nat) (mm:matZ2) (mm1:matZ2) (vf:vectorfield) (ords:orders):=
match it with
0 => (nil, nil) 
|S p => if ((longmn (size mm)(getfirstElemseq vf))&& (longmn (size (nth nil mm 0))(getsndElemseq vf)) && 
 (uniq (getfirstElemseq vf))  && (uniq (getsndElemseq vf)) && (norep ords) && (compijCvd vf mm))
   then
match mm1, vf, ords with
|nil , _ , _ => (vf, ords)
|a::nil , _ , nil => let k:= (firstElem1 j a) in
                        let colk1:= (getcol k mm) in
                             if (k < size a) 
				then  if ((canAddCvd (size mm) (size (nth nil mm 0)) i k vf) == true)
				    	then if ((canAddOrders i 0 colk1 nil) == true)
                                                 then ((vf ++ ((i, k)::nil)),  (addOrders i 0 colk1 nil))
                                             else (genDvfOrders p i (j+1) mm (a::nil) vf nil)
					else (genDvfOrders p i (j+1) mm (a::nil) vf nil)
				else (vf, nil)
|a::nil ,_,e::f => let k := (firstElem1 j a) in
	              let colk1 := (getcol k mm) in
                             if (k < size a)  
				then if ((canAddCvd (size mm) (size (nth nil mm 0)) i k vf) == true) 
					then if ((canAddOrders i 0 colk1 (e::f)) == true)
		      				then  (vf ++((i, k)::nil), (addOrders i 0 colk1 (e::f)))
						else (genDvfOrders p i (j+1) mm (a::nil) vf (e::f))
					else (genDvfOrders p i (j+1) mm (a::nil) vf (e::f))
				else (vf, (e::f)) 
|a::b, _, nil => let k := (firstElem1 j a) in
                    let colk1 := (getcol k mm) in
                    if (k < size a) 
	              then if ((canAddCvd (size mm)(size (nth nil mm 0)) i k vf) == true)
		         then if ((canAddOrders i 0 colk1 nil) == true)
                              then (genDvfOrders p (i+1) 0 mm b (addcvd i k vf) (addOrders i 0 colk1 nil)) 
                             else (genDvfOrders p i (j+1) mm (a::b) vf nil)		    
                       else (genDvfOrders p i (j+1) mm (a::b) vf nil)
		else (genDvfOrders p (i+1) 0 mm (b) vf nil)
|a::b, _, e::f => let k := (firstElem1 j a) in
	             let colk1 := (getcol k mm)	in
                  if (k < size a) 
		         then if ((canAddCvd (size mm) (size (nth nil mm 0)) i k vf) ==true)
		            then if ((canAddOrders i 0 colk1 (e::f)) == true)
		  	        then (genDvfOrders p (i+1) 0 mm b (addcvd i k vf) (addOrders i 0 colk1 ((e::f)))) 
			        else (genDvfOrders p i (j+1) mm (a::b) vf (e::f))
	 	            else (genDvfOrders p i (j+1) mm (a::b) vf (e::f))
	              else (genDvfOrders p (i+1) 0 mm (b) vf (e::f))				
end
else (nil, nil)
end.


Functional Scheme genDvfOrders_ind := Induction for genDvfOrders Sort Prop.



End vectors.




Section genDvfOrd.


Definition genDvf (ss: matZ2):=
match ss with
|nil => nil
|a::b => let dvforders:= (genDvfOrders ((size ss)*(size a)+1) 0 0 (a::b) (a::b) nil nil) in 
           if (prop_cat dvforders.2) then (dvforders.1) else nil

end.


Definition genOrders (ss: matZ2):=
match ss with
|nil => nil
|a::b => let ords:= (snd (genDvfOrders ((size ss)*(size a)+1) 0 0 (a::b) (a::b) nil nil)) in
            if (prop_cat ords) then ords else nil
end.

End genDvfOrd.

Close Scope ring_scope.


(*TO GET THE REDUCED MATRIX*)

Section matrix_reduced.


(*to order a dvf according to lengthMax1 (higher to lower) *)
(*the dvf will be ordered according to the maximum lengths of the orders which start with the first
 component of every vector*)

Section REORDERDVF.


Definition insert_sort_g (ord:orders)(s : seq (nat * nat)) := insert_sort glMax s ord.


(*this function gives us an ordered dvf taking into account the maximum lengths*)
Definition dvford (ss:matZ2):= insert_sort_g (genOrders ss)(genDvf ss).


End REORDERDVF.



(* hasta aqui cambiado*)

Section REORDERMATRIX.

(*-- "REORDER THE ROWS AND COLUMNS of the matrix according to a vector"*)
(*-- return a vector formed by the given sequence adding the elements <= k that aren't in the given vector
-- example: fill 5 [4,2,5]  ->[4,2,5,3,1]*)
Definition fill s n:= s ++ (filter [pred i | i \notin s] (iota 0 n)).


(*-- reorder the rows of the matrix according to the sequence
-- the length of this vector is the same that the number of rows of the matrix*)
Definition take_rows_s (s:seq nat) (M: matZ2):= map (nth nil M) s.


(*-- reorder the columns of a row (of a matrix) according to the sequence
-- It's used when you want reorder the columns of a matrix, the following function*)
Definition take_row_s (s:seq nat)(l:seq Z2):=
  map (nth 0%R l) s.



(*-- reorder the columns of the matrix according to the sequence
-- k is an iterator which starts with the size of ss*)
Definition take_columns_s (s:seq nat) (M: matZ2):= 
  map (take_row_s s) M.


Definition reorderM (s1 s2: seq nat)(M: matZ2):=
  (take_columns_s s2 (take_rows_s s1 M)).

(*-- "Reorder the rows and columns of the matrix according to the calculated vector field"

-- Reorder the matrix with the information of the vector field
-- example: [(1,3),(2,4),(4,5)]  -> [1,2,4] fill [1,2,4,3,5] reordernarfilas
--                               -> [3,4,5]          [3,4,5,1,2] reordenarcolumnas
-- reorder the rows of the matrix with the vector to reorder rows
-- and reorder the columns of the matrix with the vector to reorder columns*)

Definition reorderM_dvf (dvf:vectorfield)(M: matZ2):=
  reorderM (fill (getfirstElemseq dvf) (size M))
     (fill (getsndElemseq dvf) (size (head [::] M))) M.



End REORDERMATRIX.



Section GETMATRIXREDUCED.

Variable m1:nat.
Variable m: matZ2.

Definition epsilonseq := ulsubseqmx m1 m1 m.
Definition betaseq := drsubseqmx m1 m1 m.
Definition phiseq := ursubseqmx m1 m1 m.
Definition psiseq := dlsubseqmx m1 m1 m.

(* -- Given the matrix ordered and m1 the number of vectors of the dvf, return the reduced matrix*)

Definition getMatrixReduced := 
match m with
|nil => nil
|a::b => let lm := (size (a::b)) in
         let n := (size a) in
          if ((m1 == lm) || (m1 == n))
            then nil
            else subseqmx betaseq (mulseqmx (mulseqmx psiseq 
	        (cfast_invmx m1 epsilonseq)) phiseq)
end.



Definition getMatrixReduced_2 := 
match m with
|nil => nil
|a::b => let lm := (size (a::b)) in
         let n := (size a) in
          if ((m1 == lm) || (m1 == n))
            then nil
            else if (const_seqmx (lm - m1) m1 0) == phiseq 
                 then betaseq
                 else 
subseqmx betaseq (mulseqmx (mulseqmx psiseq 
	        (invmx_seqmx m1 epsilonseq)) phiseq)
end.


End GETMATRIXREDUCED.


Definition matrixReduced (m:matZ2) :=
match m with
|nil => nil
|_ => let cvd := dvford m in
      getMatrixReduced (size cvd) (reorderM_dvf cvd m)
end.


End matrix_reduced.




End definitions.

