Require Import ssreflect ssrbool ssrfun eqtype ssrnat seq div fintype.
Require Import finfun bigop prime binomial ssralg finset fingroup finalg.
Require Import perm zmodp matrix mxalgebra. 
Require Import inverse_nilpotent.

Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

(* This file is useful to be able to apply the BPL lemma in my particular case.
 One of the conditions which are necessary to apply the BPL is that:
    (H0 * P1) is nilpotent *)

(*Definitions of upper_triangular, lower_triangular and upper_triangular_up_to_k

 Lemmas proved about the relation between a lower triangular matrix and another upper
 triangular matrix.
   lower_triangular_trmx: lower_triangular A -> upper_triangular A^T 

 
 (*An upper triangular matrix is invertible*)
 Lemma upper_triangular_invertible : (upper_triangular A) -> A \in unitmx.
 


(*The product of an upper triangular matrix up to 0, A with
     an upper triangular matrix up to k, B is an upper triangular matrix up
     to k.+1*)

  Lemma upper_triangular_mulmx_AB n k (A B:'M[F]_n) :
    upper_triangular_up_to_k 0 A -> 
    upper_triangular_up_to_k k B ->
    upper_triangular_up_to_k k.+1 (A *m B).

(*If a matrix A is an upper triangular matrix then (A-Id) is an upper 
  triangular matrix up to 0*)

  Lemma upper_triangular_minus_1 n (A:'M[F]_n):
    upper_triangular A -> upper_triangular_up_to_k 0 (A - 1%:M). 


Lemma upper_triangular_pot_matrix n k (A:'M[F]_n) :
  upper_triangular_up_to_k 0 A -> 
  upper_triangular_up_to_k k (pot_matrix A k.+1). 

(*If A: M_n is an upper triangular matrix up to 0 then the power of the matrix A
  (n.+1) times is the null matrix*)
  Lemma upper_triangular_pot_matrix_n n (A:'M[F]_n) :
    upper_triangular_up_to_k 0 A -> (pot_matrix A n.+1) = 0. 


*)


Open Local Scope ring_scope.

Section Triangular.


Variable F : fieldType.


Definition upper_triangular n (A:'M[F]_(n)) :=
  forall (i j : 'I_n), j <= i -> A i j = (i == j)%:R.

Definition lower_triangular n (A:'M[F]_(n)) :=
  forall (i j : 'I_n), i <= j -> A i j = (i == j)%:R.

Lemma lower_triangular_trmx n (A:'M[F]_(n)) :
  lower_triangular A -> upper_triangular A^T. 
Proof.
rewrite /lower_triangular /upper_triangular => H i j Hij.
by rewrite mxE (H _ _ Hij) (eq_sym i j).
Qed.



Definition upper_triangular_up_to_k n k (A:'M[F]_(n)) :=
  forall (i j : 'I_n), j <= i + k -> A i j = 0.

Lemma upper_triangular_plus_1  n k (A:'M[F]_(n)):
  upper_triangular_up_to_k k.+1 A -> upper_triangular_up_to_k k A.
Proof.
rewrite /upper_triangular_up_to_k.
move => H i j H1.
have H2 : j <= i + k.+1.
  have H3 : i + k < i + k.+1. 
    by rewrite addnS ltnS.
  move : (leq_ltn_trans H1 H3) => H4.
  by rewrite leq_eqVlt; apply/orP; right.
by rewrite (H _ _ H2).
Qed.



Lemma ur_upper_triangular_plus_1 n (A:'M[F]_(1 + n)):
  upper_triangular_up_to_k (1 + n) A -> ursubmx A = 0.
Proof.
rewrite /upper_triangular_up_to_k.
move => H.
apply/matrixP => i j; rewrite !mxE.
have H1 : (rshift 1 j) <= (lshift n i) + (1 + n).
  rewrite //= [i]ord1 add0n leq_add2l.
  rewrite leq_eqVlt; apply/orP; right.
  by move : (ltn_ord j).
by rewrite (H _ _ H1).
Qed.

Lemma ul_upper_triangular_plus_1 n (A:'M[F]_(1 + n)):
  upper_triangular_up_to_k (1 + n) A -> ulsubmx A = 0.
Proof.
rewrite /upper_triangular_up_to_k.
move => H.
apply/matrixP => i j; rewrite !mxE.
have H1 : (lshift n j) <= (lshift n i) + (1 + n).
  by rewrite //= [i]ord1 [j]ord1 add0n.
by rewrite (H _ _ H1).
Qed.

Lemma dl_upper_triangular_plus_1 n (A:'M[F]_(1 + n)):
  upper_triangular_up_to_k (1 + n) A -> dlsubmx A = 0.
Proof.
rewrite /upper_triangular_up_to_k.
move => H.
apply/matrixP => i j; rewrite !mxE.
have H1 : (lshift n j) <= (rshift 1 i) + (1 + n).
  by rewrite //= [j]ord1.
by rewrite (H _ _ H1).
Qed.


Lemma dr_upper_triangular_plus_1 n (A:'M[F]_(1 + n)):
  upper_triangular_up_to_k (1 + n) A ->  
  upper_triangular_up_to_k n (drsubmx A).
Proof.
rewrite /upper_triangular_up_to_k.
move => H i j H1.
have H2 : j <= i + (1 + n).
  have H3 : i + n < i + (1 + n). 
    by rewrite [(1 + _)%N]addnC addn1 addnS ltnS.
  rewrite leq_eqVlt; apply/orP; right.
  by move : (leq_ltn_trans H1 H3).
rewrite !mxE.
have H3 : (rshift 1 j) <= (rshift 1 i) + (1 + n) by done. 
by rewrite (H _ _ H3).
Qed.

Lemma upper_triangular_up_to_dimension n (A:'M[F]_n):
   upper_triangular_up_to_k n A -> A = 0.
Proof.
move : A.
elim : n => [| n H H1].
  by move => A _; rewrite thinmx0.
move : H1.
rewrite [n.+1]/(1 + n)%N => H1 uH.
rewrite -(submxK H1) (dl_upper_triangular_plus_1 uH)
        (ul_upper_triangular_plus_1 uH)
        (ur_upper_triangular_plus_1 uH).
move : (dr_upper_triangular_plus_1 uH) => druH.
by rewrite (H _ druH) block_mx0.
Qed.

Lemma upper_triangular_k_1P  n k (A:'M[F]_n) :
  upper_triangular_up_to_k k.+1 A <->
  upper_triangular_up_to_k k A /\ 
  (forall i j :'I_n, (eq_op (@nat_of_ord n j) (addn (@nat_of_ord n i) k.+1))       -> A i j = 0).
Proof.
split.
  move => H1; split.
    apply : (upper_triangular_plus_1 H1).    
  move : H1; rewrite /upper_triangular_up_to_k => H1 i j H.
  have H2 : j<= i + k.+1.
    by rewrite leq_eqVlt; apply/orP; left.
  by rewrite (H1 _ _ H2).
rewrite /upper_triangular_up_to_k => [[H1 H2] i j H3].
move : H3; rewrite leq_eqVlt; move/orP => [HH|HH].
  by rewrite (H2 _ _ HH). 
have HH1 : j <= i + k. 
  by move : HH; rewrite addnS ltnS. 
by rewrite (H1 _ _ HH1).
Qed.

Lemma upper_triangular_mulmx n k (A:'M[F]_n) :
  upper_triangular_up_to_k k A -> 
  upper_triangular_up_to_k k.+1 (A *m A).
Proof.
elim : k => [ | k H1 H2].
   rewrite /upper_triangular_up_to_k => H i j Hij.
   rewrite !mxE /=.
   rewrite (@bigID F 0 (GRing.add_comoid F) 'I_n _ 
             (fun j0:'I_n => j0 <= i) (fun j0 => true)
             (fun j0 : 'I_n => A i j0 * A j0 j)).
   have -> : (\big[GRing.add_comoid F/0]_(i0 < n | true && (i0 <= i))
         (A i i0 * A i0 j)) = 0. 
     rewrite big1 //.
     move => i0.
     rewrite Bool.andb_true_l.
     rewrite -(addn0 i) => H1.
     by rewrite (H _ _ H1) GRing.mul0r.
   have -> : (\big[GRing.add_comoid F/0]_(i0 < n | true && ~~ (i0 <= i))
         (A i i0 * A i0 j)) = 0.
     rewrite big1 // => i0.
     rewrite Bool.andb_true_l -ltnNge => Hi0.     
     have HH : j <= i0 + 0.
       have HH1 : i +1  <= i0 by rewrite addn1. 
       by rewrite addn0 (leq_trans Hij HH1).
     by rewrite (H _ _ HH) GRing.mulr0.    
   by rewrite /= GRing.addr0.
move :  (upper_triangular_plus_1 H2) => H3.
move : (H1 H3) => HH.
rewrite upper_triangular_k_1P; split; first by done.
move => i j H. 
rewrite !mxE /=.
rewrite (@bigID F 0 (GRing.add_comoid F) 'I_n _ 
             (fun j0:'I_n => j0 <= i) (fun j0 => true)
             (fun j0 : 'I_n => A i j0 * A j0 j)).
have -> : (\big[GRing.add_comoid F/0]_(i0 < n | true && (i0 <= i))
         (A i i0 * A i0 j)) = 0. 
     rewrite big1 //.
     move => i0.
     rewrite Bool.andb_true_l.
     rewrite -(addn0 i) => HH1.
     move : H3.
     rewrite /upper_triangular_up_to_k => H3.
     have HH2 : i + 0 <= i + k. by rewrite leq_add2l. 
     by rewrite (H3 _ _ (leq_trans HH1 HH2)) GRing.mul0r.
have -> : (\big[GRing.add_comoid F/0]_(i0 < n | true && ~~ (i0 <= i))
         (A i i0 * A i0 j)) = 0.
rewrite big1 //.
move => i0.
rewrite Bool.andb_true_l -ltnNge => Hi0. 
move : H2.
rewrite /upper_triangular_up_to_k => H2.
have HH1 : j <= i0 + k.+1.
  by rewrite (eqP H) addnS ltn_add2r.
by rewrite (H2 _ _ HH1) GRing.mulr0.
by rewrite /= GRing.addr0.
Qed.

Lemma upper_triangular_mulmx_AB n k (A B:'M[F]_n) :
  upper_triangular_up_to_k 0 A -> 
  upper_triangular_up_to_k k B ->
  upper_triangular_up_to_k k.+1 (A *m B).
Proof.
elim : k => [ | k IH H1 H2].
   rewrite /upper_triangular_up_to_k => H1 H2 i j Hij.
   rewrite !mxE /=.
   rewrite (@bigID F 0 (GRing.add_comoid F) 'I_n _ 
             (fun j0:'I_n => j0 <= i) (fun j0 => true)
             (fun j0 : 'I_n => A i j0 * B j0 j)).
   have -> : (\big[GRing.add_comoid F/0]_(i0 < n | true && (i0 <= i))
         (A i i0 * B i0 j)) = 0. 
     rewrite big1 //.
     move => i0.
     rewrite Bool.andb_true_l.
     rewrite -(addn0 i) => H3.
     by rewrite (H1 _ _ H3) GRing.mul0r.
   have -> : (\big[GRing.add_comoid F/0]_(i0 < n | true && ~~ (i0 <= i))
         (A i i0 * B i0 j)) = 0.
     rewrite big1 // => i0.
     rewrite Bool.andb_true_l -ltnNge => Hi0.     
     have HH : j <= i0 + 0.
       have HH1 : i +1  <= i0 by rewrite addn1. 
       by rewrite addn0 (leq_trans Hij HH1).
     by rewrite (H2 _ _ HH) GRing.mulr0.    
   by rewrite /= GRing.addr0.
move :  (upper_triangular_plus_1 H2) => H3.
move : (IH H1 H3) => HH.
rewrite upper_triangular_k_1P; split; first by done.
move => i j H. 
rewrite !mxE /=.
rewrite (@bigID F 0 (GRing.add_comoid F) 'I_n _ 
             (fun j0:'I_n => j0 <= i) (fun j0 => true)
             (fun j0 : 'I_n => A i j0 * B j0 j)).
have -> : (\big[GRing.add_comoid F/0]_(i0 < n | true && (i0 <= i))
         (A i i0 * B i0 j)) = 0. 
     rewrite big1 //.
     move => i0.
     rewrite Bool.andb_true_l.
     rewrite -(addn0 i) => HH1.
     move : H1.
     rewrite /upper_triangular_up_to_k => H1.
     by rewrite (H1 _ _ HH1) GRing.mul0r.
have -> : (\big[GRing.add_comoid F/0]_(i0 < n | true && ~~ (i0 <= i))
         (A i i0 * B i0 j)) = 0.
rewrite big1 //.
move => i0.
rewrite Bool.andb_true_l -ltnNge => Hi0. 
move : H2.
rewrite /upper_triangular_up_to_k => H2.
have HH1 : j <= i0 + k.+1.
  by rewrite (eqP H) addnS ltn_add2r.
by rewrite (H2 _ _ HH1) GRing.mulr0.
by rewrite /= GRing.addr0.
Qed.



Lemma upper_triangular_minus_1 n (A:'M[F]_n):
  upper_triangular A -> upper_triangular_up_to_k 0 (A - 1%:M). 
Proof.
rewrite /upper_triangular /upper_triangular_up_to_k. 
move => H i j.
rewrite addn0 => Hi.
rewrite !mxE (H _ _ Hi).
case -> : (i == j); last by rewrite GRing.subr0.
rewrite /=.
have HH : 1%:R == 1%:R by done. 
move : (HH F).
rewrite -GRing.subr_eq0 => HH1.
by rewrite (eqP HH1).
Qed.

Lemma upper_triangular_pot_matrix n k (A:'M[F]_n) :
  upper_triangular_up_to_k 0 A -> 
  upper_triangular_up_to_k k (pot_matrix A k.+1). 
Proof.
elim : k => [|k H1 H2].
  by rewrite /pot_matrix mulmx1.
have -> : (pot_matrix A k.+2) = A *m (pot_matrix A k.+1) by done.
move : (H1 H2) => H3.
by exact (upper_triangular_mulmx_AB H2 H3).
Qed.

Lemma upper_triangular_pot_matrix_n n (A:'M[F]_n) :
 upper_triangular_up_to_k 0 A -> (pot_matrix A n.+1) = 0. 
Proof.
move => H.
move : (@upper_triangular_pot_matrix n n A H) => Hn.
by exact: (upper_triangular_up_to_dimension Hn).
Qed.

End Triangular.

